/*
 *
 * ICHTHYOP, a Lagrangian tool for simulating ichthyoplankton dynamics
 * http://www.ichthyop.org
 *
 * Copyright (C) IRD (Institut de Recherce pour le Developpement) 2006-2020
 * http://www.ird.fr
 *
 * Main developper: Philippe VERLEY (philippe.verley@ird.fr), Nicolas Barrier (nicolas.barrier@ird.fr)
 * Contributors (alphabetically sorted):
 * Gwendoline ANDRES, Sylvain BONHOMMEAU, Bruno BLANKE, Timothee BROCHIER,
 * Christophe HOURDIN, Mariem JELASSI, David KAPLAN, Fabrice LECORNU,
 * Christophe LETT, Christian MULLON, Carolina PARADA, Pierrick PENVEN,
 * Stephane POUS, Nathan PUTMAN.
 *
 * Ichthyop is a free Java tool designed to study the effects of physical and
 * biological factors on ichthyoplankton dynamics. It incorporates the most
 * important processes involved in fish early life: spawning, movement, growth,
 * mortality and recruitment. The tool uses as input time series of velocity,
 * temperature and salinity fields archived from oceanic models such as NEMO,
 * ROMS, MARS or SYMPHONIE. It runs with a user-friendly graphic interface and
 * generates output files that can be post-processed easily using graphic and
 * statistical software.
 *
 * To cite Ichthyop, please refer to Lett et al. 2008
 * A Lagrangian Tool for Modelling Ichthyoplankton Dynamics
 * Environmental Modelling & Software 23, no. 9 (September 2008) 1210-1214
 * doi:10.1016/j.envsoft.2008.02.005
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation (version 3 of the License). For a full
 * description, see the LICENSE file.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 *
 */

package org.previmer.ichthyop.manager;

import java.io.IOException;
import org.previmer.ichthyop.event.InitializeEvent;
import org.previmer.ichthyop.event.SetupEvent;
import org.previmer.ichthyop.io.BlockType;
import org.previmer.ichthyop.dataset.IDataset;
import org.previmer.ichthyop.io.XBlock;
import java.util.ArrayList;
import java.util.List;

/**
 *
 * @author pverley
 */
public class DatasetManager extends AbstractManager {

    final private static DatasetManager datasetManager = new DatasetManager();
    private IDataset dataset;

    public static DatasetManager getInstance() {
        return datasetManager;
    }

    private void instantiateDataset() throws Exception {

        XBlock datasetBlock = findActiveDataset();
        String className = getParameter(datasetBlock.getKey(), "class_name");
        if (datasetBlock != null) {
            try {
            dataset = (IDataset) Class.forName(className).getDeclaredConstructor().newInstance();
            } catch (Exception ex) {
                StringBuffer sb = new StringBuffer();
                sb.append("Dataset instantiation failed ==> ");
                sb.append(ex.toString());
                InstantiationException ieex = new InstantiationException(sb.toString());
                ieex.setStackTrace(ex.getStackTrace());
                throw ieex;
            }
        }
    }

    public IDataset getDataset() {
        return dataset;
    }

    public String getParameter(String datasetKey, String key) {
        return getSimulationManager().getParameterManager().getParameter(BlockType.DATASET, datasetKey, key);
    }

    private XBlock findActiveDataset() throws Exception {
        List<XBlock> list = new ArrayList<>();
        for (XBlock block : getSimulationManager().getParameterManager().getBlocks(BlockType.DATASET)) {
            if (block.isEnabled()) {
                list.add(block);
            }
        }
        if (list.isEmpty()) {
            throw new NullPointerException("Could not find any " + BlockType.DATASET.toString() + " block in the configuration file.");
        }
        if (list.size() > 1) {
            throw new IOException("Found several " + BlockType.DATASET.toString() + " blocks enabled in the configuration file. Please only keep one enabled.");
        }
        return list.get(0);
    }

    public void setupPerformed(SetupEvent e) throws Exception {
        instantiateDataset();
        getDataset().setUp();
    }

    public void initializePerformed(InitializeEvent e) throws Exception {
        getSimulationManager().getTimeManager().addNextStepListener(getDataset());
        getDataset().init();
    }
}
